﻿using System.Security.Claims;
using Infrastructure.Options;
using Microsoft.AspNetCore.Authentication;
using Microsoft.AspNetCore.Authentication.JwtBearer;
using Microsoft.IdentityModel.Tokens;

namespace Infrastructure.Extensions;

public static class AuthenticationSetup
{
    /// <summary>
    /// 配置认证服务，包含jwt认证
    /// </summary>
    /// <param name="services"></param>
    /// <param name="configuration"></param>
    /// <param name="builderOptions">自定义认证服务配置</param>
    /// <param name="configureAuthenticationOptions">配置认证选项</param>
    /// <param name="configureJwtBearerOptions">配置jwt认证选项</param>
    /// <returns></returns>
    public static IServiceCollection AddDefaultAuthentication(
        this IServiceCollection services,
        IConfiguration configuration,
        Action<AuthenticationBuilder>? builderOptions = null,
        Action<AuthenticationOptions>? configureAuthenticationOptions = null,
        Action<JwtBearerOptions>? configureJwtBearerOptions = null)
    {
        ArgumentNullException.ThrowIfNull(services);
        ArgumentNullException.ThrowIfNull(configuration);
        var audienceOptions = configuration.GetSection(AudienceOptions.Name).Get<AudienceOptions>();
        if (audienceOptions is null || !audienceOptions.IsEnable)
        {
            return services;
        }

        services.TryAddSingleton<IPostConfigureOptions<JwtBearerOptions>, JwtBearerOptionsPostConfigureOptions>();
        var key = configuration["AUDIENCE_KEY"] ?? audienceOptions.Secret;
        ArgumentException.ThrowIfNullOrEmpty(key);
        var buffer = Encoding.UTF8.GetBytes(key);
        var securityKey = new SymmetricSecurityKey(buffer);

        var tokenValidationParameters = new TokenValidationParameters()
        {
            ValidateIssuerSigningKey = true,
            IssuerSigningKey = securityKey,
            ValidIssuer = audienceOptions.Issuer,
            ValidateIssuer = true,
            ValidateAudience = true,
            ValidAudience = audienceOptions.Audience,
            ValidateLifetime = true,
            ClockSkew = TimeSpan.FromSeconds(0),
            RequireExpirationTime = true,
            RoleClaimType = ClaimTypes.Role,
            LifetimeValidator = (before, expires, token, parameters) =>
                before < DateTime.UtcNow - parameters.ClockSkew && DateTime.UtcNow < expires + parameters.ClockSkew
        };

        var builder = services.AddAuthentication(options =>
        {
            options.DefaultAuthenticateScheme = JwtBearerDefaults.AuthenticationScheme;
            options.DefaultChallengeScheme = nameof(DefaultAuthenticationHandler);
            options.DefaultForbidScheme = nameof(DefaultAuthenticationHandler);
            configureAuthenticationOptions?.Invoke(options);
        });
        builder.AddScheme<AuthenticationSchemeOptions, DefaultAuthenticationHandler>(
            nameof(DefaultAuthenticationHandler),
            options => { });
        builder.AddJwtBearer(options =>
        {
            options.TokenValidationParameters = tokenValidationParameters;
            configureJwtBearerOptions?.Invoke(options);
        });
        builderOptions?.Invoke(builder);
        return services;
    }
}